from jax import random
from collections import namedtuple

from jax_meta.metalearners.maml import MAML


ANILMetaParameters = namedtuple('ANILMetaParameters', ['encoder', 'classifier'])


class ANIL(MAML):
    def __init__(self, encoder, classifier, num_steps=5, alpha=0.01):
        super().__init__(classifier, num_steps=num_steps, alpha=alpha)
        self.encoder = encoder

    def outer_loss(self, params, state, train, test, args):
        train_features, _ = self.encoder.apply(
            params.encoder, state.encoder, train.inputs, *args
        )
        adapted_params, inner_logs = self.adapt(
            params.classifier,
            state.classifier,
            train_features,
            train.targets,
            args
        )

        test_features, state_encoder = self.encoder.apply(
            params.encoder, state.encoder, test.inputs, *args
        )
        outer_loss, (state_classifier, outer_logs) = self.loss(
            adapted_params,
            state.classifier,
            test_features,
            test.targets,
            args
        )

        state = ANILMetaParameters(
            encoder=state_encoder,
            classifier=state_classifier
        )
        return outer_loss, state, inner_logs, outer_logs

    def meta_init(self, key, inputs, *args, **kwargs):
        subkey1, subkey2 = random.split(key)

        params_encoder, state_encoder = self.encoder.init(subkey1, inputs, *args, **kwargs)
        features, _ = self.encoder.apply(params_encoder, state_encoder, inputs, *args, **kwargs)
        params_classifier, state_classifier = self.model.init(subkey2, features, *args, **kwargs)

        params = ANILMetaParameters(encoder=params_encoder, classifier=params_classifier)
        state = ANILMetaParameters(encoder=state_encoder, classifier=state_classifier)
        return (params, state)



from jax_meta.metalearners.anil import ANIL
from jax_meta.metalearners.base import MetaLearner
from jax_meta.metalearners.imaml import iMAML
from jax_meta.metalearners.lrd2 import LRD2Binary, LRD2OneVsRest
from jax_meta.metalearners.maml import MAML
from jax_meta.metalearners.meta_sgd import MetaSGD
from jax_meta.metalearners.r2d2 import R2D2


import os
import json
import shutil

from urllib.request import urlretrieve
from pathlib import Path


def download_url(url, root, filename=None):
    root = root.expanduser()
    if filename is None:
        filename = os.path.basename(url)
    filepath = root / filename

    if filepath.is_file():
        return

    root.mkdir(exist_ok=True)
    urlretrieve(url, filename=filepath)
    return filepath


def get_asset_path(*args):
    jax_meta_folder = Path(__file__).parent.parent
    assets_folder = jax_meta_folder / 'datasets' / 'assets'
    return assets_folder.joinpath(*args)


def get_asset(*args, dtype=None):
    filename = get_asset_path(*args)
    if not filename.exists():
        raise IOError(f'File not found: {filename}')

    if dtype is None:
        dtype = filename.suffix
        dtype = dtype[1:]

    if dtype == 'json':
        with open(filename, 'r') as f:
            data = json.load(f)
    else:
        raise NotImplementedError(f'Unknown type: {dtype}')
    return data


def copy_dataset_from_repository(name, destination):
    try:
        repository = get_asset('repository.json', dtype='json')
    except IOError:
        raise NotImplementedError(f'Using the dataset `{name}` in jax_meta '
            'requires a private repository of datasets, and therefore is not '
            'publicly available yet. This will be made available in a future release.')

    filename = Path(repository['datasets'][name]['local'])
    shutil.copy(filename, destination / filename.name)
    return destination / filename.name


def update_repository(origin=None):
    repo_filepath = get_asset_path('repository.json')
    if origin is None:
        repository = get_asset('repository.json', dtype='json')
        origin = Path(repository['meta']['origin'])
    shutil.copy(origin, repo_filepath)



import pytest

import numpy as np
from numpy.random import default_rng

import jax_meta.datasets.transforms.functional as F


def test_random_crop():
    # Random data
    rng = default_rng(0)
    data = rng.integers(256, size=(2, 25, 84, 84, 3), dtype=np.uint8)

    # Random crop
    rng = default_rng(1)
    data_F = F.random_crop(data, size=84, padding=8, rng=rng)
    assert data_F.shape == (2, 25, 84, 84, 3)
    assert data_F.dtype == data.dtype

    # Numpy crop
    rng = default_rng(1)  # Same RNG
    data_np = np.zeros_like(data)
    padded = np.pad(data, ((0, 0), (0, 0), (8, 8), (8, 8), (0, 0)))
    x = rng.integers(16, size=(2, 25, 1, 1, 1))
    y = rng.integers(16, size=(2, 25, 1, 1, 1))
    for i in range(2):
        for j in range(25):
            for k in range(84):
                for l in range(84):
                    for c in range(3):
                        row = x[i, j, 0, 0, 0] + k
                        col = y[i, j, 0, 0, 0] + l
                        data_np[i, j, k, l, c] = padded[i, j, row, col, c]
    np.testing.assert_equal(data_F, data_np)


def test_random_horizontal_flip():
    # Random data
    rng = default_rng(0)
    data = rng.integers(256, size=(2, 25, 84, 84, 3), dtype=np.uint8)

    # Random horizontal flip
    rng = default_rng(1)
    data_copy = np.copy(data)
    data_F = F.random_horizontal_flip(data_copy, rng=rng)
    assert data_F.shape == (2, 25, 84, 84, 3)
    np.testing.assert_equal(data_copy, data_F)  # inplace changes

    # Numpy horizontal flip
    rng = default_rng(1)  # Same RNG
    to_flip = (rng.random(size=(2, 25)) < 0.5)
    data_np = np.zeros_like(data)
    for i in range(2):
        for j in range(25):
            if to_flip[i, j]:
                data_np[i, j] = data[i, j, :, ::-1]
            else:
                data_np[i, j] = data[i, j]
    np.testing.assert_equal(data_F, data_np)



import jax.numpy as jnp
import optax
import jax

from jax import vmap, grad, tree_util, jit, lax
from functools import partial

from jax_meta.metalearners.base import MetaLearnerState
from jax_meta.metalearners.maml import MAML


class iMAML(MAML):
    def __init__(
            self,
            model,
            num_steps=5,
            alpha=0.1,
            lambda_=1.,
            regu_coef=1.,
            cg_damping=10.,
            cg_steps=5
        ):
        super().__init__(model, num_steps=num_steps, alpha=alpha)
        self.lambda_ = lambda_
        self.regu_coef = regu_coef
        self.cg_damping = cg_damping
        self.cg_steps = cg_steps

    def adapt(self, init_params, state, inputs, targets, args):
        loss_grad = grad(self.loss, has_aux=True)
        # Gradient descent with proximal regularization
        gradient_descent = lambda p, p0, g: p - self.alpha * (g + self.lambda_ * (p - p0))
        def _gradient_update(params, _):
            # Do not update the state during adaptation
            grads, (_, logs) = loss_grad(params, state, inputs, targets, args)
            params = tree_util.tree_map(gradient_descent, params, init_params, grads)
            return params, logs

        return lax.scan(
            _gradient_update,
            init_params,
            None,
            length=self.num_steps
        )

    def hessian_vector_product(self, params, state, inputs, targets, args):
        train_loss = lambda primals: self.loss(primals, state, inputs, targets, args)[0]
        _, hvp_fn = jax.linearize(grad(train_loss), params)

        def _hvp_damping(tangents):
            damping = lambda h, t: (1. + self.regu_coef) * t + h / (self.lambda_ + self.cg_damping)
            return tree_util.tree_multimap(damping, hvp_fn(tangents), tangents)

        return _hvp_damping

    def grad_outer_loss(self, params, state, train, test, args):
        @partial(vmap, in_axes=(None, None, 0, 0))
        def _grad_outer_loss(params, state, train, test):
            adapted_params, inner_logs = self.adapt(
                params, state, train.inputs, train.targets, args
            )

            # Compute the gradient wrt. the adapted parameters
            grads, (state, outer_logs) = grad(self.loss, has_aux=True)(
                adapted_params, state, test.inputs, test.targets, args)

            # Compute the meta-gradient using Conjugate Gradient
            hvp_fn = self.hessian_vector_product(
                adapted_params, state, train.inputs, train.targets, args
            )
            outer_grads, _ = jax.scipy.sparse.linalg.cg(hvp_fn, grads, maxiter=self.cg_steps)

            return outer_grads, inner_logs, outer_logs, state

        outer_grads, inner_logs, outer_logs, states = _grad_outer_loss(
            params, state, train, test)
        outer_grads = tree_util.tree_map(partial(jnp.mean, axis=0), outer_grads)
        state = tree_util.tree_map(partial(jnp.mean, axis=0), states)
        
        logs = {
            **{f'inner/{k}': v for (k, v) in inner_logs.items()},
            **{f'outer/{k}': v for (k, v) in outer_logs.items()}
        }
        return outer_grads, (state, logs)

    @partial(jit, static_argnums=(0, 5))
    def train_step(self, params, state, train, test, args):
        grads, (model_state, logs) = self.grad_outer_loss(
            params, state.model, train, test, args)

        updates, opt_state = self.optimizer.update(grads, state.optimizer, params)
        params = optax.apply_updates(params, updates)

        state = MetaLearnerState(model=model_state, optimizer=opt_state)

        return params, state, logs



import haiku as hk
import jax.numpy as jnp


class Adjust(hk.Module):
    def __init__(self, base, scale_init=1e-4, bias_init=0., name=None):
        super().__init__(name=name)
        self.base = base
        self._scale_init = scale_init
        self._bias_init = bias_init

    def __call__(self, inputs):
        # Parameters
        scale_init = hk.initializers.Constant(self._scale_init)
        scale = hk.get_parameter('scale', (), inputs.dtype, init=scale_init)

        bias_init = hk.initializers.Constant(self._bias_init)
        bias = hk.get_parameter('bias', (), inputs.dtype, init=bias_init)

        if self.base == 1:
            return inputs * scale + bias
        else:
            return inputs * (self.base ** scale) + self.base ** bias - 1



import numpy as np

from numpy.random import default_rng


def random_crop(data, size, padding, rng=default_rng()):
    x = rng.integers(2 * padding, size=data.shape[:-3] + (1, 1, 1))
    y = rng.integers(2 * padding, size=data.shape[:-3] + (1, 1, 1))
    arange = np.arange(size)
    rows = x + arange.reshape((size, 1, 1))
    cols = y + arange.reshape((1, size, 1))

    padding = (padding, padding)
    data = np.pad(data, ((0, 0),) * (data.ndim - 3) + (padding, padding, (0, 0)))
    data = np.take_along_axis(data, rows, axis=-3)
    data = np.take_along_axis(data, cols, axis=-2)
    return data


def random_horizontal_flip(data, rng=default_rng()):
    to_flip = (rng.random(size=data.shape[:-3]) < 0.5)
    data[to_flip] = np.flip(data[to_flip], axis=-2)
    return data


def normalize(data, mean, std):
    data -= mean
    data /= std
    return data


def _blend(img1, img2, ratio):
    ratio = ratio.reshape((-1,) + (1,) * (img1.ndim - 1))
    img1 *= ratio
    img1 += (1. - ratio) * img2
    img1 = np.clip(img1, 0., 1., out=img1)
    return img1


def rgb_to_grayscale(data):
    col = np.array([0.2989, 0.587, 0.114], dtype=data.dtype)
    return np.expand_dims(data.dot(col), axis=-1)


def adjust_brightness(data, factor):
    return _blend(data, np.zeros_like(data), factor)


def adjust_saturation(data, factor):
    gray = rgb_to_grayscale(data)
    return _blend(data, gray, factor)


def adjust_contrast(data, factor):
    mean_gray = np.mean(rgb_to_grayscale(data), axis=(-3, -2, -1), keepdims=True)
    return _blend(data, mean_gray, factor)


def color_jitter(data, brightness, contrast, saturation, rng=default_rng()):
    order = np.argsort(rng.random(size=(3,) + data.shape[:-3]), axis=0)
    brightness = rng.uniform(1. - brightness, 1. + brightness, size=data.shape[:-3])
    contrast = rng.uniform(1. - contrast, 1. + contrast, size=data.shape[:-3])
    saturation = rng.uniform(1. - saturation, 1. + saturation, size=data.shape[:-3])
    for transform in order:
        data[transform == 0] = adjust_brightness(data[transform == 0], brightness[transform == 0])
        data[transform == 1] = adjust_contrast(data[transform == 1], contrast[transform == 1])
        data[transform == 2] = adjust_saturation(data[transform == 2], saturation[transform == 2])
    return data



import numpy as np
import os
import pickle

from collections import defaultdict

from jax_meta.datasets.base import MetaDataset
from jax_meta.utils.data import download_url


class LEOMetaDataset(MetaDataset):
    name = 'leo'
    zip_url = 'http://storage.googleapis.com/leo-embeddings/embeddings.zip'
    shape = (640,)

    def __init__(
            self,
            root,
            batch_size,
            shots=5,
            ways=5,
            test_shots=None,
            size=None,
            crop='center',
            split='train',
            seed=0,
            download=False
        ):
        self.crop = crop
        super().__init__(root, batch_size, shots=shots, ways=ways,
            test_shots=test_shots, size=size, split=split, seed=seed,
            download=download)
        self.load_data()

    def load_data(self):
        if self._data is None:
            arrays, labels2indices = [], defaultdict(list)
            offset = 0
            for filename in self.split_filenames:
                with open(filename, 'rb') as f:
                    data = pickle.load(f, encoding='latin')
                    arrays.append(data['embeddings'])

                    for i, key in enumerate(data['keys']):
                        _, class_name, _ = str(key).split('-')
                        labels2indices[class_name].append(i + offset)

                    offset += data['embeddings'].shape[0]
            self._data = np.concatenate(arrays, axis=0)
            self._labels2indices = dict((k, np.asarray(v))
                for (k, v) in labels2indices.items())
        return self

    @property
    def split_filenames(self):
        folder = self.folder / 'embeddings' / self.subname / self.crop
        return tuple(folder / f'{split}_embeddings.pkl' for split in self.splits)

    def _check_integrity(self):
        return all(filename.exists() for filename in self.split_filenames)

    def download(self):
        import zipfile

        if self._check_integrity():
            return

        # Download dataset
        download_url(self.zip_url, self.folder)

        # Extract dataset
        filename = self.folder / os.path.basename(self.zip_url)
        if not filename.with_suffix('').is_dir():
            with zipfile.ZipFile(filename, 'r') as f:
                f.extractall(self.folder)
        filename.unlink(missing_ok=True)


class LEOMiniImagenet(LEOMetaDataset):
    subname = 'miniImageNet'


class LEOTieredImagenet(LEOMetaDataset):
    subname = 'tieredImageNet'



import numpy as np
import jax.numpy as jnp
import os
import wandb
import warnings


def slurm_infos():
    return {
        'slurm/job_id': os.getenv('SLURM_JOB_ID'),
        'slurm/job_user': os.getenv('SLURM_JOB_USER'),
        'slurm/job_partition': os.getenv('SLURM_JOB_PARTITION'),
        'slurm/cpus_per_node': os.getenv('SLURM_JOB_CPUS_PER_NODE'),
        'slurm/num_nodes': os.getenv('SLURM_JOB_NUM_NODES'),
        'slurm/nodelist': os.getenv('SLURM_JOB_NODELIST'),
        'slurm/cluster_name': os.getenv('SLURM_CLUSTER_NAME'),
        'slurm/array_task_id': os.getenv('SLURM_ARRAY_TASK_ID')
    }


def table_from_array(array):
    data = [(value,) + index for (index, value) in np.ndenumerate(array)]
    columns = ['data'] + [f'x_{i}' for i in range(array.ndim)]
    return wandb.Table(data=data, columns=columns)


def to_wandb(logs, prefix=None, keys=None):
    converted_logs = {}
    prefix = '' if (prefix is None) else f'{prefix}/'
    for key, value in logs.items():
        if (keys is not None) and (key not in keys):
            continue

        if np.ndim(value) == 0:
            converted_logs[f'{prefix}{key}'] = value

        elif isinstance(value, (np.ndarray, jnp.DeviceArray)):
            converted_logs[f'{prefix}{key}'] = table_from_array(value)

        else:
            warnings.warn(f'Unknown type {type(value)} for key: {key}')
    return converted_logs



import jax.numpy as jnp


def accuracy(outputs, targets):
    return jnp.mean(jnp.argmax(outputs, axis=-1) == targets)


def binary_accuracy(logits, targets):
    return jnp.mean((logits > 0).astype(targets.dtype) == targets)



import haiku as hk
import math

from jax import nn


class ConvBlock(hk.Module):
    def __init__(self, channels, name=None):
        super().__init__(name=name)
        self.channels = channels

    def __call__(self, inputs, is_training):
        outputs = inputs
        outputs = hk.Conv2D(self.channels, kernel_shape=3,
            stride=1, with_bias=True, name='conv')(outputs)
        outputs = hk.BatchNorm(create_scale=True, create_offset=True,
            decay_rate=0.9, name='norm')(outputs, is_training)
        outputs = nn.relu(outputs)
        outputs = hk.max_pool(outputs, 2, 2, padding='VALID')
        return outputs


class Conv4(hk.Module):
    def __init__(self, num_filters=64, normalize_outputs=False, name=None):
        super().__init__(name=name)
        self.num_filters = num_filters
        self.normalize_outputs = normalize_outputs

    def __call__(self, inputs, is_training):
        outputs = inputs
        outputs = ConvBlock(self.num_filters, name='layer1')(outputs, is_training)
        outputs = ConvBlock(self.num_filters, name='layer2')(outputs, is_training)
        outputs = ConvBlock(self.num_filters, name='layer3')(outputs, is_training)
        outputs = ConvBlock(self.num_filters, name='layer4')(outputs, is_training)
        outputs = outputs.reshape(inputs.shape[:-3] + (-1,))
        normalization = math.sqrt(outputs.shape[-1]) if self.normalize_outputs else 1.
        return outputs / normalization



import jax.numpy as jnp

from jax import nn


def nll_loss(log_likelihoods, targets):
    losses = jnp.take_along_axis(log_likelihoods, targets[..., None], axis=-1)
    return -jnp.squeeze(losses, axis=-1)


def cross_entropy(logits, targets):
    log_likelihoods = nn.log_softmax(logits, axis=-1)
    return nll_loss(log_likelihoods, targets)


def binary_cross_entropy(logits, targets):
    # See: https://www.tensorflow.org/api_docs/python/tf/nn/sigmoid_cross_entropy_with_logits
    return nn.relu(logits) - logits * targets + nn.softplus(-jnp.abs(logits))



import numpy as np
import math

from abc import ABC
from pathlib import Path
from numpy.random import default_rng
from collections import namedtuple


Dataset = namedtuple('Dataset', ['inputs', 'targets', 'infos'])


class MetaDataset(ABC):
    def __init__(
            self,
            root,
            batch_size,
            shots=5,
            ways=5,
            test_shots=None,
            size=None,
            split='train',
            seed=0,
            download=False
        ):
        self.root = Path(root).expanduser()
        self.folder = self.root / self.name
        self.batch_size = batch_size
        self.shots = shots
        self.ways = ways
        self.test_shots = shots if (test_shots is None) else test_shots
        self.size = size

        self.splits = split.split('+')
        assert all(split in ['train', 'val', 'test'] for split in self.splits)

        if download:
            self.download()

        self.seed = seed
        self.reset()

        self._data = None
        self._labels2indices = None
        self._labels = None
        self._num_classes = None

    def reset(self):
        self.rng = default_rng(self.seed)
        self.num_samples = 0
        return self

    @property
    def data(self):
        return self._data

    @property
    def labels2indices(self):
        return self._labels2indices

    @property
    def labels(self):
        if self._labels is None:
            self._labels = sorted(self.labels2indices.keys())
        return self._labels

    @property
    def num_classes(self):
        if self._num_classes is None:
            self._num_classes = len(self.labels)
        return self._num_classes

    def get_indices(self):
        total_shots = self.shots + self.test_shots

        class_indices = np.zeros((self.batch_size, self.ways), dtype=np.int_)
        indices = np.zeros((self.batch_size, self.ways, total_shots), dtype=np.int_)
        targets = np.zeros((self.batch_size, self.ways), dtype=np.int_)

        for idx in range(self.batch_size):
            class_indices[idx] = self.rng.choice(self.num_classes, size=(self.ways,), replace=False)
            targets[idx] = self.rng.permutation(self.ways)

            for way in range(self.ways):
                label = self.labels[class_indices[idx, way]]
                indices[idx, way] = self.rng.choice(self.labels2indices[label], size=(total_shots,), replace=False)

        return class_indices, indices, targets

    def transform(self, data):
        return data

    def __len__(self):
        if self.size is None:
            raise RuntimeError('The dataset has no length because it is infinite.')
        return self.size

    def __iter__(self):
        while (self.size is None) or (self.num_samples < self.size):
            class_indices, indices, targets = self.get_indices()
            data = self.transform(self.data[indices])

            train = Dataset(
                inputs=data[:, :, :self.shots].reshape((self.batch_size, -1) + data.shape[3:]),
                targets=targets.repeat(self.shots, axis=1),
                infos={'labels': class_indices, 'indices': indices[..., :self.shots]}
            )
            test = Dataset(
                inputs=data[:, :, self.shots:].reshape((self.batch_size, -1) + data.shape[3:]),
                targets=targets.repeat(self.test_shots, axis=1),
                infos={'labels': class_indices, 'indices': indices[..., self.shots:]}
            )

            self.num_samples += 1
            yield {'train': train, 'test': test}

    def download(self):
        pass

    @property
    def dummy_input(self):
        return self.transform(np.zeros((1,) + self.data.shape[1:], dtype=np.uint8))

    def flatten(self, num_epochs=1):
        return FlattenDataset(self, num_epochs=num_epochs)


class FlattenDataset:
    def __init__(self, dataset, num_epochs=1):
        self.dataset = dataset
        self.num_epochs = num_epochs
        self.batch_size = dataset.batch_size
        self.size = len(dataset.data)
        self._labels = None
        self.reset()

    @property
    def labels(self):
        if self._labels is None:
            self._labels = np.zeros((self.size,), dtype=np.int_)
            for i, label in enumerate(self.dataset.labels):
                self._labels[self.dataset.labels2indices[label]] = i
        return self._labels

    def reset(self):
        self.dataset.reset()
        self.rng = self.dataset.rng

    def __len__(self):
        return self.num_epochs * math.ceil(self.size / self.batch_size)

    def __iter__(self):
        for _ in range(self.num_epochs):
            indices = self.rng.permutation(self.size)

            for i in range(0, self.size, self.batch_size):
                slice_ = indices[i:i + self.batch_size]
                inputs = self.dataset.transform(self.dataset.data[slice_])
                targets = self.labels[slice_]
                yield inputs, targets



import numpy as np
import pickle

from jax_meta.datasets.miniimagenet import MiniImagenet


class TieredImagenet(MiniImagenet):
    name = 'tieredimagenet'
    shape = (84, 84, 3)

    def load_data(self):
        if self._data is None:
            arrays, labels2indices = [], {}
            offset = 0
            for data_filename, labels_filename in self.split_filenames:
                with open(data_filename, 'rb') as f:
                    data = np.load(f)
                    arrays.append(data['images'])

                with open(labels_filename, 'rb') as f:
                    data = pickle.load(f)
                    labels = np.asarray(data['labels'])

                    unique_labels = np.unique(labels)
                    labels2indices.update({label: offset + np.where(labels == label)[0]
                        for label in unique_labels})

                    offset += labels.shape[0]
            self._data = np.concatenate(arrays, axis=0)
            self._labels2indices = labels2indices
        return self

    @property
    def split_filenames(self):
        filenames = {
            'train': ('train_images.npz', 'train_labels.pkl'),
            'val': ('val_images.npz', 'val_labels.pkl'),
            'test': ('test_images.npz', 'test_labels.pkl')
        }
        return tuple((self.folder / filenames[split][0],
                      self.folder / filenames[split][1])
            for split in self.splits)

    def _check_integrity(self):
        for data_filename, labels_filename in self.split_filenames:
            if (not data_filename.is_file()) or (not labels_filename.is_file()):
                return False
        return True



import pytest
import haiku as hk
import optax
import jax.numpy as jnp

from jax import random

from jax_meta.metalearners.base import MetaLearner, MetaLearnerState
from jax_meta.metalearners.maml import MAML


@pytest.fixture
def model():
    def network(inputs, _):
        return hk.Linear(5)(inputs)
    return hk.without_apply_rng(hk.transform_with_state(network))


def test_maml(model):
    metalearner = MAML(model, num_steps=1, alpha=0.1)
    assert isinstance(metalearner, MetaLearner)


def test_maml_init(model):
    metalearner = MAML(model, num_steps=1, alpha=0.1)
    optimizer = optax.adam(1e-3)

    key = random.PRNGKey(0)
    inputs = jnp.zeros((2, 3))
    _, state = metalearner.init(key, optimizer, inputs, True)

    assert isinstance(state, MetaLearnerState)



import jax.numpy as jnp
import optax

from abc import ABC, abstractmethod
from jax import jit, grad, tree_util, vmap
from functools import partial
from collections import namedtuple


MetaLearnerState = namedtuple('MetaLearnerState', ['model', 'optimizer'])


class MetaLearner(ABC):
    def __init__(self):
        self._optimizer = None

    @abstractmethod
    def loss(self, params, state, inputs, targets, args):
        pass

    @abstractmethod
    def adapt(self, params, state, inputs, targets, args):
        pass

    @abstractmethod
    def meta_init(self, key, *args, **kwargs):
        pass

    def step(self, params, state, train, test, *args):
        return self.train_step(params, state, train, test, args)

    def outer_loss(self, params, state, train, test, args):
        adapted_params, inner_logs = self.adapt(
            params, state, train.inputs, train.targets, args
        )
        outer_loss, (state, outer_logs) = self.loss(
            adapted_params, state, test.inputs, test.targets, args
        )
        return (outer_loss, state, inner_logs, outer_logs)

    @partial(jit, static_argnums=(0, 5))
    def train_step(self, params, state, train, test, args):
        outer_loss_grad = grad(self.batch_outer_loss, has_aux=True)
        grads, (model_state, logs) = outer_loss_grad(
            params, state.model, train, test, args)

        # Apply gradient descent
        updates, opt_state = self.optimizer.update(grads, state.optimizer, params)
        params = optax.apply_updates(params, updates)

        state = MetaLearnerState(model=model_state, optimizer=opt_state)

        return params, state, logs

    @property
    def optimizer(self):
        if self._optimizer is None:
            raise ValueError(f'`{self:s}` contains no optimizer. To train the'
                             'model, you must call the `init` function.')
        return self._optimizer

    def batch_outer_loss(self, params, state, train, test, args):
        outer_loss = vmap(self.outer_loss, in_axes=(None, None, 0, 0, None))
        outer_losses, states, inner_logs, outer_logs = outer_loss(
            params, state, train, test, args
        )
        state = tree_util.tree_map(partial(jnp.mean, axis=0), states)

        logs = {
            **{f'inner/{k}': v for (k, v) in inner_logs.items()},
            **{f'outer/{k}': v for (k, v) in outer_logs.items()}
        }
        return jnp.mean(outer_losses), (state, logs)

    def init(self, key, optimizer, *args, **kwargs):
        self._optimizer = optimizer
        params, model_state = self.meta_init(key, *args, **kwargs)
        state = MetaLearnerState(
            model=model_state,
            optimizer=self.optimizer.init(params)
        )
        return params, state

    def evaluate(self, params, state, dataset, *args):
        if dataset.size is None:
            raise RuntimeError('The dataset for evaluation must be finite, '
                'got an infinite dataset. You must set the `size` argument '
                'when creating the dataset.')

        outer_loss = jit(self.batch_outer_loss, static_argnums=(4,))
        results = None
        for i, batch in enumerate(dataset.reset()):
            _, (_, logs) = outer_loss(params, state.model, batch['train'], batch['test'], args)
            logs = tree_util.tree_map(lambda arr: jnp.mean(arr, axis=0), logs)

            if results is None:
                results = logs
            else:
                online_mean = lambda mean, arr: mean + ((arr - mean) / (i + 1))
                results = tree_util.tree_map(online_mean, results, logs)

        return results



import numpy as np
import pickle

from jax_meta.datasets.base import MetaDataset
from jax_meta.utils.data import copy_dataset_from_repository
import jax_meta.datasets.transforms.functional as F


class FC100(MetaDataset):
    name = 'fc100'
    shape = (32, 32, 3)

    def __init__(
            self,
            root,
            batch_size,
            shots=5,
            ways=5,
            test_shots=None,
            size=None,
            split='train',
            seed=0,
            data_augmentation=None,
            download=False
        ):
        super().__init__(root, batch_size, shots=shots, ways=ways,
            test_shots=test_shots, size=size, split=split, seed=seed,
            download=download)
        self.load_data()
        if self.data_augmentation is None:
            self._data_augmentation = ('train' in self.splits)
        else:
            self._data_augmentation = data_augmentation
        self._mean = np.array([129.37731888, 124.10583864, 112.47758569]) / 255.
        self._std = np.array([68.20947949, 65.43124043, 70.45866994]) / 255.

    def load_data(self):
        if self._data is None:
            arrays, labels2indices = [], {}
            offset = 0
            for filename in self.split_filenames:
                with open(filename, 'rb') as f:
                    data = pickle.load(f, encoding='latin1')
                    arrays.append(data['data'])
                    labels = np.asarray(data['labels'])

                    unique_labels = np.unique(labels)
                    labels2indices.update({label: offset + np.where(labels == label)[0]
                        for label in unique_labels})

                    offset += labels.shape[0]
            self._data = np.concatenate(arrays, axis=0)
            self._labels2indices = labels2indices
        return self

    def transform(self, data):
        if self._data_augmentation:
            data = F.random_horizontal_flip(data, rng=self.rng)
            data = data.astype(np.float32) / 255.
            data = F.color_jitter(data, brightness=0.4, contrast=0.4, saturation=0.4, rng=self.rng)
            data = F.normalize(data, self._mean, self._std)
            data = F.random_crop(data, size=32, padding=4, rng=self.rng)

        else:
            data = data.astype(np.float32) / 255.
            data = F.normalize(data, self._mean, self._std)

        return data

    @property
    def split_filenames(self):
        return tuple(self.folder / f'FC100_{split}.pickle' for split in self.splits)

    def _check_integrity(self):
        return all(filename.exists() for filename in self.split_filenames)

    def download(self):
        import zipfile

        if self._check_integrity():
            return

        self.folder.mkdir(exist_ok=True)
        filename = copy_dataset_from_repository(self.name, self.folder)

        with zipfile.ZipFile(filename, 'r') as f:
            f.extractall(self.folder)
        filename.unlink(missing_ok=True)



from jax_meta.datasets.base import MetaDataset, FlattenDataset

from jax_meta.datasets.cifarfs import CIFARFS
from jax_meta.datasets.fc100 import FC100
from jax_meta.datasets.leo import LEOMiniImagenet, LEOTieredImagenet
from jax_meta.datasets.miniimagenet import MiniImagenet
from jax_meta.datasets.omniglot import Omniglot
from jax_meta.datasets.tieredimagenet import TieredImagenet



import os
import enum

from typing import Optional
from dataclasses import dataclass
from simple_parsing.helpers import Serializable, encode

from jax_meta import datasets


class Dataset(enum.Enum):
    omniglot = 'omniglot'
    miniimagenet = 'miniimagenet'
    tieredimagenet = 'tieredimagenet'
    cifarfs = 'cifarfs'
    fc100 = 'fc100'

    @property
    def _cls(self):
        classes = {
            Dataset.omniglot: datasets.Omniglot,
            Dataset.miniimagenet: datasets.MiniImagenet,
            Dataset.tieredimagenet: datasets.TieredImagenet,
            Dataset.cifarfs: datasets.CIFARFS,
            Dataset.fc100: datasets.FC100
        }
        if self not in classes:
            raise ValueError(f'Unknown dataset: `{self}`')

        return classes[self]


@encode.register(Dataset)
def encode_dataset(dataset):
    return dataset.value


@dataclass
class DataArguments(Serializable):
    # data folder
    folder: Optional[str]

    # dataset name
    dataset: Dataset = Dataset.miniimagenet

    # number of classes per task (N in "N-way", default: %(default)s)
    ways: int = 5

    # number of training examples per class (k in "k-shot", default: %(default)s)
    shots: int = 5

    # number of test examples per class
    test_shots: int = 15

    # random seed
    seed: int = 1

    def __post_init__(self):
        if self.folder is None:
            if os.getenv('SLURM_TMPDIR') is not None:
                self.folder = os.path.join(os.getenv('SLURM_TMPDIR'), 'data')
            else:
                raise ValueError(f'Invalid value of `folder`: {self.folder}. '
                    '`folder` must be a valid folder.')
        os.makedirs(self.folder, exist_ok=True)

        if self.test_shots <= 0:
            self.test_shots = self.shots



import numpy as np
import haiku as hk

from typing import Mapping


def save(filename, **trees):
    data = {}
    for prefix, tree in trees.items():
        if isinstance(tree, Mapping):
            tree = hk.data_structures.to_haiku_dict(tree)
            for module_name, name, value in hk.data_structures.traverse(tree):
                data[f'{prefix}/{module_name}/{name}'] = value
        else:
            data[prefix] = tree

    np.savez(filename, **data)


def load(filename, **kwargs):
    data = {}
    f = open(filename, 'rb') if isinstance(filename, str) else filename
    results = np.load(f, **kwargs)

    for key in results.files:
        prefix, delimiter, name = key.rpartition('/')
        if delimiter:
            prefix, _, module_name = prefix.partition('/')
            if prefix not in data:
                data[prefix] = {}
            if module_name not in data[prefix]:
                data[prefix][module_name] = {}
            data[prefix][module_name][name] = results[key]
        else:
            data[name] = results[key]

    for prefix, tree in data.items():
        if isinstance(tree, dict):
            data[prefix] = hk.data_structures.to_haiku_dict(tree)
    
    if isinstance(filename, str):
        f.close()
    del results

    return data



import numpy as np
import os
import pickle

from jax_meta.datasets.fc100 import FC100


class CIFARFS(FC100):
    name = 'cifarfs'
    shape = (32, 32, 3)

    @property
    def split_filenames(self):
        return tuple(self.folder / f'CIFAR_FS_{split}.pickle' for split in self.splits)



import jax.numpy as jnp
import haiku as hk
import math

from jax import nn, random
from collections import namedtuple

from jax_meta.utils.losses import cross_entropy
from jax_meta.utils.metrics import accuracy
from jax_meta.metalearners.base import MetaLearner
from jax_meta.modules.r2d2 import Adjust


R2D2MetaParameters = namedtuple('R2D2MetaParameters', ['encoder', 'adjust'])


class R2D2(MetaLearner):
    def __init__(self, encoder, adjust=None, num_ways=5, lambda_=0.1):
        super().__init__()

        if adjust is None:
            @hk.without_apply_rng
            @hk.transform
            def adjust(inputs):
                return Adjust(1)(inputs)

        self.encoder = encoder
        self.adjust = adjust
        self.num_ways = num_ways
        self.lambda_ = lambda_

    def loss(self, params, state, linear_params, inputs, targets, args):
        features, state = self.model.apply(params.encoder, state, inputs, *args)
        features = R2D2.append_ones(features)

        outputs = jnp.matmul(features, linear_params)
        logits = self.adjust.apply(params.adjust, outputs)
        loss = jnp.mean(cross_entropy(logits, targets))
        logs = {
            'loss': loss,
            'accuracy': accuracy(logits, targets),
        }
        return loss, (state, logs)

    def adapt(self, params, state, inputs, targets, args):
        num_samples = inputs.shape[0]
        I = jnp.eye(num_samples)

        features, _ = self.model.apply(params.encoder, state, inputs, *args)
        features = features / math.sqrt(num_samples)
        features = R2D2.append_ones(features)

        A = jnp.matmul(features, features.T) + self.lambda_ * I
        b = nn.one_hot(targets, self.num_ways) / math.sqrt(num_samples)
        solution = jnp.linalg.solve(A, b)

        return (jnp.matmul(features.T, solution), {})

    def meta_init(self, key, *args, **kwargs):
        subkey1, subkey2 = random.split(key)
        encoder, encoder_state = self.model.init(subkey1, *args, **kwargs)

        features = self.model.apply(encoder, encoder_state, *args, **kwargs)
        adjust = self.adjust.init(subkey2, features)

        params = R2D2MetaParameters(encoder=encoder, adjust=adjust)
        return params, encoder_state

    @staticmethod
    def append_ones(features):
        ones = jnp.ones(features.shape[:-1] + (1,), dtype=features.dtype)
        return jnp.append(features, ones, axis=-1)



import jax.numpy as jnp

from jax import vmap, lax, nn

from jax_meta.utils.losses import binary_cross_entropy, cross_entropy
from jax_meta.utils.metrics import binary_accuracy, accuracy
from jax_meta.metalearners.base import MetaLearner


class LRD2Binary(MetaLearner):
    def __init__(self, encoder, lambda_=0.1, num_iterations=10):
        super().__init__()
        self.encoder = encoder
        self.lambda_ = lambda_
        self.num_iterations = num_iterations

    def loss(self, params, state, linear_params, inputs, targets, args):
        features, state = self.model.apply(params, state, inputs, *args)
        logits = jnp.matmul(features, linear_params)
        loss = jnp.mean(binary_cross_entropy(logits, targets))
        logs = {
            'loss': loss,
            'accuracy': binary_accuracy(logits, targets),
        }
        return loss, (state, logs)

    def adapt(self, params, state, inputs, targets, args):
        features, _ = self.model.apply(params, state, inputs, *args)
        def _irls_step(linear_params, _):
            num_samples = inputs.shape[0]
            I = jnp.eye(num_samples)

            # Predictions
            logits = jnp.matmul(features, linear_params)
            preds = nn.sigmoid(logits)
            s = preds * (1 - preds)

            A = jnp.matmul(features, features.T * s) + self.lambda_ * I
            b = logits + (targets - preds) / s
            solution = jnp.linalg.solve(A, b)

            logs = {
                'loss': jnp.mean(binary_cross_entropy(logits, targets)),
                'accuracy': binary_accuracy(logits, targets),
            }
            return jnp.dot(features.T * s, solution), logs

        init_params = jnp.zeros((features.shape[-1],), dtype=features.dtype)
        return lax.scan(_irls_step, init_params, None, length=self.num_iterations)

    def meta_init(self, key, *args, **kwargs):
        return self.model.init(key, *args, **kwargs)


class LRD2OneVsRest(LRD2Binary):
    def __init__(
            self,
            encoder,
            lambda_=0.1,
            num_iterations=10,
            num_ways=5
        ):
        super().__init__(
            encoder,
            lambda_=lambda_,
            num_iterations=num_iterations
        )
        self.num_ways = num_ways

    def loss(self, params, state, linear_params, inputs, targets, args):
        features, state = self.model.apply(params, state, inputs, *args)
        logits = jnp.matmul(features, linear_params)
        loss = jnp.mean(cross_entropy(logits, targets))
        logs = {
            'loss': loss,
            'accuracy': accuracy(logits, targets),
        }
        return loss, (state, logs)

    def adapt(self, params, state, inputs, targets, args):
        features, _ = self.model.apply(params, state, inputs, *args)
        one_hot_targets = nn.one_hot(targets, self.num_ways)
        def _irls_step(linear_params, _):
            num_samples = inputs.shape[0]
            I = jnp.eye(num_samples)

            # Predictions
            logits = jnp.dot(features, linear_params)
            preds = nn.sigmoid(logits)
            s = preds * (1 - preds)

            A = jnp.matmul(features, features.T) * jnp.expand_dims(s.T, axis=1) + self.lambda_ * I
            b = logits + (one_hot_targets - preds) / s
            solution = jnp.linalg.solve(A, b.T)

            logs = {
                'loss': jnp.mean(cross_entropy(logits, targets)),
                'accuracy': accuracy(logits, targets),
            }

            new_params = vmap(jnp.dot)(
                features.T * jnp.expand_dims(s.T, axis=1), solution)
            return new_params.T, logs

        init_params = jnp.zeros((features.shape[-1], self.num_ways), dtype=features.dtype)
        return lax.scan(_irls_step, init_params, None, length=self.num_iterations)



import jax.numpy as jnp

from jax import grad, tree_util, lax
from collections import namedtuple

from jax_meta.metalearners.maml import MAML


MetaSGDMetaParameters = namedtuple('MetaSGDMetaParameters', ['model', 'alpha'])


class MetaSGD(MAML):
    def adapt(self, params, state, inputs, targets, args):
        loss_grad = grad(self.loss, has_aux=True)
        gradient_descent = lambda p, a, g: p - a * g  # Gradient descent
        def _gradient_update(params, _):
            # Do not update the state during adaptation
            grads, (_, logs) = loss_grad(params.model, state, inputs, targets, args)
            params = tree_util.tree_multimap(
                gradient_descent,
                params.model,
                params.alpha,
                grads
            )
            return params, logs

        return lax.scan(_gradient_update, params, None, length=self.num_steps)

    def meta_init(self, key, *args, **kwargs):
        model, model_state = self.model.init(key, *args, **kwargs)
        alpha = tree_util.tree_map(lambda x: jnp.full_like(x, self.alpha), model)
        params = MetaSGDMetaParameters(model=model, alpha=alpha)
        return params, model_state



import jax.numpy as jnp
import haiku as hk
import math

from jax import nn


class ResidualBlock(hk.Module):
    def __init__(self, channels, stride, name=None):
        super().__init__(name=name)
        self.channels = channels
        self.stride = stride

    def __call__(self, inputs, is_training):
        residuals = outputs = inputs

        outputs = hk.Conv2D(self.channels, kernel_shape=3,
            stride=1, with_bias=False, name='conv1')(outputs)
        outputs = hk.BatchNorm(create_scale=True, create_offset=True,
            decay_rate=0.9, name='norm1')(outputs, is_training)
        outputs = nn.leaky_relu(outputs, negative_slope=0.1)

        outputs = hk.Conv2D(self.channels, kernel_shape=3,
            stride=1, with_bias=False, name='conv2')(outputs)
        outputs = hk.BatchNorm(create_scale=True, create_offset=True,
            decay_rate=0.9, name='norm2')(outputs, is_training)
        outputs = nn.leaky_relu(outputs, negative_slope=0.1)

        outputs = hk.Conv2D(self.channels, kernel_shape=3,
            stride=1, with_bias=False, name='conv3')(outputs)
        outputs = hk.BatchNorm(create_scale=True, create_offset=True,
            decay_rate=0.9, name='norm3')(outputs, is_training)

        residuals = hk.Conv2D(self.channels, kernel_shape=1,
            stride=1, with_bias=False, name='downsample_conv')(residuals)
        residuals = hk.BatchNorm(create_scale=True, create_offset=True,
            decay_rate=0.9, name='downsample_norm')(residuals, is_training)

        outputs = outputs + residuals
        outputs = nn.leaky_relu(outputs, negative_slope=0.1)
        outputs = hk.max_pool(outputs, self.stride, self.stride, padding='VALID')

        return outputs


class ResNet12(hk.Module):
    def __init__(self, num_filters=64, normalize_outputs=False, name=None):
        super().__init__(name=name)
        self.num_filters = num_filters  # Unused
        self.normalize_outputs = normalize_outputs

    def __call__(self, inputs, is_training):
        outputs = inputs
        outputs = ResidualBlock(64, 2, name='block1')(outputs, is_training)
        outputs = ResidualBlock(160, 2, name='block2')(outputs, is_training)
        outputs = ResidualBlock(320, 2, name='block3')(outputs, is_training)
        outputs = ResidualBlock(640, 2, name='block4')(outputs, is_training)
        outputs = jnp.mean(outputs, axis=(1, 2))
        normalization = math.sqrt(outputs.shape[-1]) if self.normalize_outputs else 1.
        return outputs / normalization



import jax.numpy as jnp
import haiku as hk


class RelationNetwork(hk.nets.MLP):
    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        num_inputs = inputs.shape[-2]

        left = jnp.expand_dims(inputs, axis=-2).repeat(num_inputs, axis=-2)
        right = jnp.expand_dims(inputs, axis=-3).repeat(num_inputs, axis=-3)
        concatenated = jnp.concatenate([left, right], axis=-1)

        outputs =  super().__call__(concatenated)
        return jnp.mean(outputs, axis=-2)



import os
import sys
import shutil
import warnings

from io import open
from pathlib import Path
from setuptools import setup, find_packages
from setuptools.command.install import install

here = Path(__file__).parent.resolve()

sys.path.insert(0, str(here / 'jax_meta'))
from version import VERSION

class InstallCommand(install):
    def run(self):
        super().run()

        if 'JAX_META_REPOSITORY' in os.environ:
            repository = Path(os.environ['JAX_META_REPOSITORY'])
            if not repository.exists():
                warnings.warn(f'Repository not found (ignoring): {repository}')
                return

            destination = Path(self.install_lib) / 'jax_meta' / 'datasets' / 'assets'
            shutil.copy(repository, destination / 'repository.json')

# Get the long description from the README file
with open(here / 'README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='jax_meta',
    version=VERSION,
    description='Collection of meta-learning algorithms in Jax',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    author='Tristan Deleu',
    author_email='tristan.deleu@gmail.com',
    url='https://github.com/tristandeleu/jax-meta-learning',
    keywords=['meta-learning', 'jax', 'few-shot', 'few-shot learning'],
    packages=find_packages(exclude=['data', 'contrib', 'docs', 'tests', 'examples']),
    install_requires=[
        'jax',
        'jaxlib',
        'tqdm',
        'simple-parsing',
        'Pillow>=7.0.0',
        'dm-haiku',
        'optax'
    ],
    package_data={'jax_meta': ['jax_meta/datasets/assets/*']},
    include_package_data=True,
    cmdclass={'install': InstallCommand},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
    ],
)



import jax.numpy as jnp

from jax import grad, tree_util, lax

from jax_meta.utils.losses import cross_entropy
from jax_meta.utils.metrics import accuracy
from jax_meta.metalearners.base import MetaLearner


class MAML(MetaLearner):
    def __init__(self, model, num_steps=5, alpha=0.1):
        super().__init__()
        self.model = model
        self.num_steps = num_steps
        self.alpha = alpha

    def loss(self, params, state, inputs, targets, args):
        logits, state = self.model.apply(params, state, inputs, *args)
        loss = jnp.mean(cross_entropy(logits, targets))
        logs = {
            'loss': loss,
            'accuracy': accuracy(logits, targets),
        }
        return loss, (state, logs)

    def adapt(self, init_params, state, inputs, targets, args):
        loss_grad = grad(self.loss, has_aux=True)
        gradient_descent = lambda p, g: p - self.alpha * g  # Gradient descent
        def _gradient_update(params, _):
            # Do not update the state during adaptation
            grads, (_, logs) = loss_grad(params, state, inputs, targets, args)
            params = tree_util.tree_map(gradient_descent, params, grads)
            return params, logs

        return lax.scan(
            _gradient_update,
            init_params,
            None,
            length=self.num_steps
        )

    def meta_init(self, key, *args, **kwargs):
        return self.model.init(key, *args, **kwargs)



import numpy as np
import json
import concurrent.futures

from glob import glob
from PIL import Image
from collections import defaultdict

import jax_meta.datasets.transforms.functional as F
from jax_meta.datasets.base import MetaDataset
from jax_meta.utils.data import download_url, get_asset


class Omniglot(MetaDataset):
    name = 'omniglot'
    url = 'https://raw.githubusercontent.com/brendenlake/omniglot/master/python'
    filenames = ['images_background', 'images_evaluation']
    shape = (28, 28, 1)

    def __init__(
            self,
            root,
            batch_size,
            shots=5,
            ways=5,
            test_shots=None,
            size=None,
            split='train',
            seed=0,
            download=False
        ):
        super().__init__(root, batch_size, shots=shots, ways=ways,
            test_shots=test_shots, size=size, split=split, seed=seed,
            download=download)
        self.load_data()

    def load_data(self):
        if self._data is None:
            arrays, labels2indices = [], {}
            offset = 0
            for data_filename, labels_filename in self.split_filenames:
                with open(data_filename, 'rb') as f:
                    data = np.load(f)
                    arrays.append(data)

                with open(labels_filename, 'r') as f:
                    labels = json.load(f)

                    labels2indices.update({label: offset + np.array(indices)
                        for label, indices in labels.items()})

                    offset += data.shape[0]
            arrays = np.concatenate(arrays, axis=0)

            # Add class augmentations
            data, labels2indices_aug, num_samples = [], {}, arrays.shape[0]
            for k in [0, 1, 2, 3]:  # Rotation by 0, 90, 180, 270
                data.append(np.rot90(arrays, k=k, axes=(1, 2)))
                labels2indices_aug.update({
                    f'{key}_{k * 90}': values + num_samples * k
                    for (key, values) in labels2indices.items()
                })

            self._data = np.concatenate(data, axis=0)
            self._labels2indices = labels2indices_aug
        return self

    def transform(self, data):
        return 1. - data.astype(np.float32) / 255.

    @property
    def split_filenames(self):
        filenames = {
            'train': ('train_images.npy', 'train_labels.json'),
            'val': ('val_images.npy', 'val_labels.json'),
            'test': ('test_images.npy', 'test_labels.json')
        }
        return tuple((self.folder / filenames[split][0],
                      self.folder / filenames[split][1])
            for split in self.splits)

    def _check_integrity(self):
        for data_filename, labels_filename in self.split_filenames:
            if (not data_filename.is_file()) or (not labels_filename.is_file()):
                return False
        return True

    def download(self, max_workers=8):
        import zipfile
        import shutil

        if self._check_integrity():
            return

        for filename in self.filenames:
            zip_filename = f'{filename}.zip'
            filename = self.folder / zip_filename

            if filename.is_file():
                continue

            download_url(f'{self.url}/{zip_filename}', self.folder, zip_filename)
            with zipfile.ZipFile(filename, 'r') as f:
                f.extractall(self.folder)

        def load_image(filename, size=28):
            image = Image.open(filename, mode='r').convert('L')
            return image.resize((size, size), Image.BILINEAR)

        for split in ['train', 'val', 'test']:
            data = get_asset(self.name, f'{split}.json', dtype='json')
            characters = []
            for folder in ['background', 'evaluation']:
                characters.extend([
                    (f'{alphabet}/{character}', filename)
                    for (alphabet, characters) in data[folder].items()
                    for character in characters
                    for filename in glob(str(self.folder / f'images_{folder}' / alphabet / character / '*.png'))
                ])
            characters = sorted(characters)

            images = np.zeros((len(characters),) + self.shape, dtype=np.uint8)
            with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
                filenames = [filename for _, filename in characters]
                for i, image in enumerate(executor.map(load_image, filenames)):
                    images[i, ..., 0] = image

            labels = defaultdict(list)
            for i, (label, _) in enumerate(characters):
                labels[label].append(i)

            with open(self.folder / f'{split}_images.npy', 'wb') as f:
                np.save(f, images)

            with open(self.folder / f'{split}_labels.json', 'w') as f:
                json.dump(labels, f)

        for filename in self.filenames:
            filename = self.folder / f'{filename}.zip'
            shutil.rmtree(filename.with_suffix(''))
            filename.unlink(missing_ok=True)
